''' Source Dedicated Server console handling functions '''

import os
import sys
import es
import gamethread
import traceback
import srcds

def excepter(e_type, e_value, e_traceback):
    dbgmsg(''.join(traceback.format_exception(e_type, e_value, e_traceback)))

def dbgmsg(text):
    sys.stderr.write(str(text)+"\n")

class FDError(RuntimeError):
    ''' An error with FDcatcher '''
    pass

class FDCatcher(object):
    ''' A class to manipulate file descriptors '''
    def __init__(self, fd):
        ''' FDCatcher(fd)  -  returns a file descriptor manipulating object
    for file descriptor fd '''
        self.originalfd = fd
        self.oldfd = os.dup(fd)
        self._state = None
    def getoutput(self):
        ''' Returns a new file descriptor from which one can read the output
of the original file descriptor '''
        if self._state:
            raise FDError, 'Cannot get output before releasing old'
        readfd, writefd = os.pipe()
        self._state = readfd, writefd
        os.close(self.originalfd)
        os.dup2(writefd, self.originalfd)
        return readfd
    def getinput(self):
        ''' Returns a new file descriptor to which one can write the input
for the original file descriptor '''
        if self._state:
            raise FDError, 'Cannot get input before releasing old'
        readfd, writefd = os.pipe()
        os.close(self.originalfd)
        os.dup2(readfd, self.originalfd)
        #os.close(readfd)
        self._state = readfd, writefd
        return writefd
    def release(self):
        ''' Release the input or output file descriptor and return to normal
behavior '''
        if not self._state:
            raise FDError, 'Nothing to release'
        for fd in self._state:
            os.close(fd)
        os.close(self.originalfd)
        os.dup2(self.oldfd, self.originalfd)
        self._state = None
    def __int__(self):
        ''' Returns the original fd given on initialization '''
        return self.originalfd
    def __del__(self):
        ''' x.__del__() <=> del x '''
        os.close(self.oldfd)

stdout_fn = sys.__stdout__.fileno()
if stdout_fn >= 0:
    stdout_catcher = FDCatcher(sys.__stdout__.fileno())
else:
    stdout_catcher = None

class _RequestHandler(list):
    def append(self, obj):
        list.append(self, obj)
        self._processQueue()
    def _processQueue(self):
        if not srcds.is_loaded:
            es.load('srcds')
            gamethread.delayed(0, self._processQueue)
        elif self:
            self._execute()
    def _execute(self):
        obj = self.pop(0)
        obj.doit(self._processQueue)

class _CommandHandler(object):
    def __init__(self):
        self.handlers = {}
    def add(self, handler):
        thisid = id(handler)
        self.handlers[thisid] = handler
        return thisid
    def delete(self, thisid):
        del self.handlers[thisid]
    def start(self, thisid):
        self.handlers[thisid]._start_capture()
    def end(self, thisid):
        self.handlers[thisid]._end_capture()
    def __contains__(self, item):
        return item in self.handlers


_requests = _RequestHandler()
_concmd_handlers = _CommandHandler()

        
class _Concmd(object):
    def __init__(self, command, callback, args, kw, bufsize):
        self.command = command
        self.callback = callback
        self.args = args
        self.kw = kw
        self.bufsize = bufsize
    def doit(self, process):
        self.callid = _concmd_handlers.add(self)
        es.ServerCommand('_start_console_capture %d'%self.callid)
        es.ServerCommand(self.command)
        es.ServerCommand('_end_console_capture %d'%self.callid)
        self.process = process
    def _start_capture(self):
        self.fd = stdout_catcher.getoutput()
    def _end_capture(self):
        _concmd_handlers.delete(self.callid)
        self.call(self.process)
    def call(self, process):
        try:
            output = os.read(self.fd, self.bufsize)
            stdout_catcher.release()
            self.fd = None
            self.callback(output, *self.args, **self.kw)
        except Exception, e:
            dbgmsg(traceback.format_exc())
        process()

def get_concmd_output(command, callback, args = (), kw = None,
                      bufsize = 1024):
    ''' Get output of a console command

Parameters:
    command     The console command to execute
    callback    A callable Python object that must accept at least
                1 parameter, the first parameter will be the output
                in console saved in a string
    args        Extra parameters for the callback function
    kw          Extra keyvalue parameters for the callback function
    bufsize     The maximum number of characters of the output to store

Returns:
    None
    '''
    if not kw:
        kw = {}
    if not isinstance(args, tuple):
        args = args,
    if not callable(callback):
        raise TypeError, '%s is not callable'%repr(callback)
    _requests.append(_Concmd(command, callback, args, kw, bufsize))

def get_console_output(command, cargs = (), ckw = None,
                       bufsize = 1024):
    ''' Get console output of a Python function

Parameters:
    command     The Python callable to be called
    cargs       The parameters for the function
    ckw         Extra keyvalue parameters for the function
    bufsize     The maximum number of characters of the output to be stored

Returns:
    None
    '''
    if not ckw:
        ckw = {}
    if not isinstance(cargs, tuple):
        cargs = cargs,
    if not callable(command):
        raise TypeError, '%s is not callable'%repr(command)
    fd = stdout_catcher.getoutput()
    the_exception = None
    try:
        return_value = command(*cargs, **ckw)
    except Exception, e:
        the_exeception, e_value, e_tb = sys.exc_info()
    try:
        output = os.read(fd, bufsize)
        stdout_catcher.release()
    except Exception, e:
        dbgmsg(traceback.format_exc())
    if the_exception:
        raise the_exception, e_value, e_tb
    return return_value, output

def dummy(*args, **kw):
    pass
